/*
 *  Copyright 2009 RISC                            
 *  Copyright 2009 Karbin                          
 *  Distributed under the terms of the MIT License.
 */

module sample.chicago.ChicagoApp;

import sample.chicago.crowd;

import tango.io.Stdout;
import tango.stdc.stringz;
import tango.math.Math;

import sample.HordeApp;

void main()
{
	auto chicago = new ChicagoApp();

	// Open window
	chicago.openWindow();

	// Initialize application
	if (!chicago.init())
	{
		Stderr("Failed to initialize Horde3D.").newline;
		return;
	}

	// Run the application
	chicago.run();

	// Release the engine
	chicago.release();
}

public class ChicagoApp : HordeApp
{
	// Engine objects
	private ResHandle pipeRes, lightMatRes, logoMatRes, forwardPipeRes, deferredPipeRes, skyBoxRes;
	private NodeHandle particleSys, sky;

	private bool freeze;
	private float animTime;
	private float weight;

	private float x, y, z, rx, ry; // Viewer position and orientation
	private float velocity; // Velocity for movement
	private CrowdSim crowdSim; 

	public this()
	{
		x = 15;
		y = 3;
		z = 20;
		rx = -10;
		ry = 60;

		animTime = 0.0f;
		weight = 1.0f;
		velocity = 10.0f;
	}

	public override bool init()
	{
		if (!super.init())
		{
			return false;
		}

		// Add resources
		// Pipelines
		forwardPipeRes = Horde3D.addResource(ResourceTypes.Pipeline, "forward.pipeline.xml", 0);
		deferredPipeRes = Horde3D.addResource(ResourceTypes.Pipeline, "deferred.pipeline.xml", 0);
		// Logo
		logoMatRes = Horde3D.addResource(ResourceTypes.Material, "logo.material.xml", 0);
		// Shader for deferred shading
		ResHandle lightMatRes = Horde3D.addResource(ResourceTypes.Material, "light.material.xml", 0);
		// Environment
		ResHandle envRes = Horde3D.addResource(ResourceTypes.SceneGraph, "platform.scene.xml", 0);
		// Skybox
		ResHandle skyBoxRes = Horde3D.addResource(ResourceTypes.SceneGraph, "skybox.scene.xml", 0);
		
		// Load resources
		Horde3DUtils.loadResourcesFromDisk("Content");

		// Add scene nodes
		// Add camera
		cam = Horde3D.addCameraNode(RootNode, "Camera", forwardPipeRes);
		//Horde3D.setNodeParami(cam, CameraNodeParams.OcclusionCulling, 1);
		// Add environment
		NodeHandle env = Horde3D.addNodes(RootNode, envRes);
		Horde3D.setNodeTransform(env, 0, 0, 0, 0, 0, 0, 0.23f, 0.23f, 0.23f);
		// Add skybox
		NodeHandle sky = Horde3D.addNodes(RootNode, skyBoxRes);
		Horde3D.setNodeTransform(sky, 0, 0, 0, 0, 0, 0, 210, 50, 210);
		// Add light source
		NodeHandle light = Horde3D.addLightNode(RootNode, "Light1", lightMatRes, "LIGHTING", "SHADOWMAP");
		Horde3D.setNodeTransform(light, 0, 20, 50, -30, 0, 0, 1, 1, 1);
		Horde3D.setNodeParamf(light, LightNodeParams.Radius, 200);
		Horde3D.setNodeParamf(light, LightNodeParams.FOV, 90);
		Horde3D.setNodeParami(light, LightNodeParams.ShadowMapCount, 3);
		Horde3D.setNodeParamf(light, LightNodeParams.ShadowSplitLambda, 0.9f);
		Horde3D.setNodeParamf(light, LightNodeParams.ShadowMapBias, 0.001f);
		Horde3D.setNodeParamf(light, LightNodeParams.Col_R, 0.9f);
		Horde3D.setNodeParamf(light, LightNodeParams.Col_G, 0.7f);
		Horde3D.setNodeParamf(light, LightNodeParams.Col_B, 0.75f);

		crowdSim = new CrowdSim( "Content" );
		
		return true;
	}

	protected override void render()
	{
		if (isShowingFPS())
		{
			char[] text = "Weigth: " ~ Float.format(new char[32], weight);
			Horde3DUtils.showText(toStringz(text), 0, 0.91f, 0.03f, 0, fontMatRes);
		}

		// Set camera parameters
		Horde3D.setNodeTransform(cam, x, y, z, rx, ry, 0, 1, 1, 1);
		// Show logo
		Horde3D.showOverlay(0.75f, 0, 0, 0, 1, 0, 1, 0, 1, 0.2f, 1, 1, 0.75f, 0.2f, 0, 1, 7, logoMatRes);
	}

	protected override void update()
	{
		super.update();
		
		float curVel = velocity / getFPS();

		if (wasKeyPressed(SDLK_ESCAPE))
		{
			setRunning(false);
		}

		if (wasKeyPressed(SDLK_SPACE))
		{
			toggleFreeze();
		}

		if (wasKeyPressed(SDLK_F3))
		{
			togglePipeline();
		}

		if (wasKeyPressed(SDLK_F7))
		{
			toggleDebugViewMode();
		}

		if (wasKeyPressed(SDLK_F8))
		{
			toggleWireframeMode();
		}

		if (wasKeyPressed(SDLK_F9))
		{
			toggleShowFPS();
		}

		if (isKeyDown(SDLK_LSHIFT))
		{
			curVel *= 5;
		}

		if (isKeyDown(SDLK_w))
		{
			// Move forward
			x -= sin(degToRad(ry)) * cos(-degToRad(rx)) * curVel;
			y -= sin(-degToRad(rx)) * curVel;
			z -= cos(degToRad(ry)) * cos(-degToRad(rx)) * curVel;
		}

		if (isKeyDown(SDLK_s))
		{
			// Move backward
			x += sin(degToRad(ry)) * cos(-degToRad(rx)) * curVel;
			y += sin(-degToRad(rx)) * curVel;
			z += cos(degToRad(ry)) * cos(-degToRad(rx)) * curVel;
		}

		if (isKeyDown(SDLK_a))
		{
			// Strafe left
			x += sin(degToRad(ry - 90)) * curVel;
			z += cos(degToRad(ry - 90)) * curVel;
		}
		if (isKeyDown(SDLK_d))
		{
			// Strafe right
			x += sin(degToRad(ry + 90)) * curVel;
			z += cos(degToRad(ry + 90)) * curVel;
		}
		if (isKeyDown(SDLK_1))
		{
			// Change blend weight
			weight += 2 / getFPS();
			if (weight > 1)
			{
				weight = 1;
			}
		}
		if (isKeyDown(SDLK_2))
		{
			// Change blend weight
			weight -= 2 / getFPS();
			if (weight < 0)
			{
				weight = 0;
			}
		}
		
		if (!freeze)
		{
			animTime += 1.0f / getFPS();
			
			crowdSim.update(getFPS());
			
		}
	}

	protected override void mouseMoved(int xrel, int yrel)
	{
		// Look left/right
		ry -= cast (float) xrel * 0.01 * 30;

		// Loop up/down but only in a limited range
		final float newX = cast (float) yrel * 0.01 * 30;
		rx += isInvertMouse() ? -newX : newX;
		
		clamp(rx, -90, 90);
	}
	
	private void toggleFreeze()
	{
		freeze = !freeze;
	}
	
	private void togglePipeline()
	{
		if (Horde3D.getNodeParami(cam, CameraNodeParams.PipelineRes) == forwardPipeRes)
		{
			Horde3D.setNodeParami(cam, CameraNodeParams.PipelineRes, deferredPipeRes);
		}
		else
		{
			Horde3D.setNodeParami(cam, CameraNodeParams.PipelineRes, forwardPipeRes);
		}
	}
	
	private void clamp(inout float value, float min, float max)
	in
	{
		assert (max > min);
	}
	body
	{
		if(value > max)
		{
			value = max;
		}
		else if(value < min)
		{
			value = min;
		}
	}
	
	private float degToRad(float f)
	{
		return f * (PI / 180.0f);
	}
}
